#include "RdmaSalEndpoint.hpp"
#include "MessageFormats.hpp"
RdmaSalEndpoint::RdmaSalEndpoint(struct rdma_cm_id *id, struct ibv_cq *completionQueue, int sendQueueSize,
                                 int recvQueueSize, int sendMsgSize, int recvMsgSize, rocksdb::DB *db)
    : RdmaEndpoint(id, completionQueue, sendQueueSize, recvQueueSize, sendMsgSize, recvMsgSize), _db(db)
{
}

void RdmaSalEndpoint::processSendCompletion(struct ibv_wc *data)
{
    /*means data has been send to other side we can use this buffer*/
    _sendBuffers->push((void *)data->wr_id);
}

int RdmaSalEndpoint::sendMessage(const char *buffer, uint32_t size)
{
    if (size > _sendMsgSize)
        return -1;
    void *sendBuffer = nullptr;
    _sendBuffers->pop(sendBuffer);
    if (sendBuffer == nullptr)
        return -1;
    memcpy(sendBuffer, buffer, size);
    return rdma_post_send(_cm_id, sendBuffer, sendBuffer, size, _sendMr, 0);
}

void RdmaSalEndpoint::processRecvCompletion(struct ibv_wc *data)
{
    char *request = new char[data->byte_len];
    memcpy(request, (void *)data->wr_id, data->byte_len);
    struct SalRequestHeader *req = (struct SalRequestHeader *)request;
    rdma_post_recv(_cm_id, (void *)data->wr_id, (void *)data->wr_id, _recvMsgSize, _recvMr);
    if (req->type == RequestType::DELETE)
        processDelete(req);
    if (req->type == RequestType::GET)
        processGet(req);
    if (req->type == RequestType::PUT)
        processPut(req);
    delete[] request;
}

void RdmaSalEndpoint::processDelete(struct SalRequestHeader *req)
{
    rocksdb::Status s = _db->Delete(rocksdb::WriteOptions(), {(char *)req + SalRequestHeaderSize, req->keySize});
    void *sendBuf = nullptr;
    _sendBuffers->pop(sendBuf);
    if (sendBuf == nullptr)
    {
        return;
    }
    SalResponseHeader *response = (SalResponseHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->status = ResponseStatus::FAILURE;
    response->id = req->id;
    if (s.ok())
    {
        response->status = ResponseStatus::SUCCESS;
    }
    rdma_post_send(_cm_id, sendBuf, sendBuf, SalResponseHeaderSize, _sendMr, 0);
}

void RdmaSalEndpoint::processGet(struct SalRequestHeader *req)
{
    std::string value;
    rocksdb::Status s = _db->Get(rocksdb::ReadOptions(), {(char *)req + SalRequestHeaderSize, req->keySize}, &value);
    void *sendBuf = nullptr;
    _sendBuffers->pop(sendBuf);
    if (sendBuf == nullptr)
    {
        return;
    }
    SalResponseHeader *response = (SalResponseHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->status = ResponseStatus::FAILURE;
    response->id = req->id;
    if (s.ok())
    {
        response->status = ResponseStatus::SUCCESS;
        response->valueSize = value.size();
        memcpy(response + SalResponseHeaderSize, value.c_str(), value.size());
    }
    rdma_post_send(_cm_id, sendBuf, sendBuf, SalRequestHeaderSize + value.size(), _sendMr, 0);
}

void RdmaSalEndpoint::processPut(struct SalRequestHeader *req)
{
    rocksdb::Status s = _db->Put(rocksdb::WriteOptions(), {(char *)req + SalRequestHeaderSize, req->keySize},
                                 {(char *)req + SalRequestHeaderSize + req->keySize, req->valueSize});
    void *sendBuf = nullptr;
    _sendBuffers->pop(sendBuf);
    if (sendBuf != nullptr)
    {
        return;
    }
    SalResponseHeader *response = (SalResponseHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->status = ResponseStatus::FAILURE;
    response->id = req->id;
    if (s.ok())
        response->status = ResponseStatus::FAILURE;
    rdma_post_send(_cm_id, sendBuf, sendBuf, SalResponseHeaderSize, _sendMr, 0);
}
